import json
from typing import Any, Dict
from django.apps import apps
from django.http import HttpResponse
from django.utils.decorators import method_decorator
from django.views.decorators.csrf import csrf_exempt
from django.views.generic import View


@method_decorator(csrf_exempt, name="dispatch")
class GoogleCloudTaskView(View):
    """Endpoints for on-demand and periodic tasks"""

    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self.tasks = self._get_available_tasks()

    def _get_available_tasks(self):
        task_manager = apps.get_app_config("django_gcp").task_manager
        all_tasks = task_manager.on_demand_tasks.copy()
        all_tasks.update(task_manager.periodic_tasks.copy())
        return all_tasks

    def post(self, request, task_name, *args, **kwargs):
        """Receive POST request containing task data and execute the task"""
        try:
            task_class = self.tasks[task_name]
        except KeyError:
            status = 404
            result = {"error": f"Task {task_name} not found", "available_tasks": list(self.tasks)}
            return self._prepare_response(status=status, payload=result)

        output, status = task_class().execute(request_body=request.body)
        if status == 200:
            result = {"result": output}
        else:
            result = {"error": output}

        return self._prepare_response(status=status, payload=result)

    def _prepare_response(self, status: int, payload: Dict[str, Any]):
        return HttpResponse(status=status, content=json.dumps(payload), content_type="application/json")


class GoogleCloudSubscriberTaskView(GoogleCloudTaskView):
    """Endpoints for subscriber tasks"""

    def _get_available_tasks(self):
        return apps.get_app_config("django_gcp").task_manager.subscriber_tasks.copy()



import json
from google.cloud.logging_v2.handlers import StructuredLogHandler


class GoogleStructuredLogsHandler(StructuredLogHandler):
    """A logging handler which structures logs so they're readable in Cloud Run log tabs

    The StructuredLogHandler from Google is used as a basis but the httpRequest field is
    injected to preserve request context (empty by default otherwise), as per the requirements
    at:
    https://cloud.google.com/logging/docs/reference/v2/rest/v2/LogEntry#HttpRequest

    TODO The httpRequest object could have much richer data than this, which should be accessible
    from `record.request` or by adding more advanced middleware. PRs are welcome.

    """

    def format(self, record):
        """Override the default formatter to inject httpRequest information"""

        # Inject http request status to the python logger for logs from django.server
        if (record.module == "basehttp") and record.status_code is not None:
            request_method, request_url, protocol = record.args[0].split(" ")
            httpRequest = {
                "requestMethod": request_method,
                "requestUrl": request_url,
                "protocol": protocol,
                "status": record.status_code,
            }
            # Without overriding the entire format() method, injecting this protected value
            # is the only way of adding httpRequest data to the formatted output
            # pylint: disable-next=protected-access
            record._http_request_str = json.dumps(httpRequest, ensure_ascii=False, cls=self._json_encoder_cls)
            # Override the django.server default log message to remove the quotation marks
            record.msg = "%s %s %s"

        return super().format(record)



from gcp_pilot.base import GoogleCloudPilotAPI
from gcp_pilot.scheduler import CloudScheduler as BaseCloudScheduler


class CloudScheduler(BaseCloudScheduler):
    """Override the location setter on the CloudScheduler

    GCP pilot sets default location by accessing AppEngine.
    If a project has no AppEngine instance, a NotFound exception is caught and silently handled, returning None.

    In GCP Pilot's AppEngineBasedService mixin (used by both CloudScheduler and CloudTasks), the _set_location
    method is overridden to force use of the default value (which in this case is erroneously None), and any
    other value is disallowed.

    GCP Pilot should instead use the project description to find the defaults (or otherwise determine the region
    for cloud scheduler and tasks).

    We must therefore override the override (!!) back to use the method from the original base class,
    and set the location by explicitly passing it as a parameter.
    """

    def _set_location(self, location: str = None):
        return GoogleCloudPilotAPI._set_location(self, location)  # pylint: disable=protected-access
